from collections import OrderedDict
from decimal import Decimal, InvalidOperation

import arrow  # type: ignore

from isoduration.parser.exceptions import (
    IncorrectDesignator,
    NoTime,
    OutOfDesignators,
    UnknownToken,
    UnparseableValue,
)
from isoduration.parser.util import (
    is_letter,
    is_number,
    is_time,
    is_week,
    parse_designator,
)
from isoduration.types import DateDuration, Duration, TimeDuration


def parse_datetime_duration(duration_str: str, sign: int) -> Duration:
    try:
        duration: arrow.Arrow = arrow.get(duration_str)
    except (arrow.ParserError, ValueError):
        raise UnparseableValue(f"Value could not be parsed as datetime: {duration_str}")

    return Duration(
        DateDuration(
            years=sign * duration.year,
            months=sign * duration.month,
            days=sign * duration.day,
        ),
        TimeDuration(
            hours=sign * duration.hour,
            minutes=sign * duration.minute,
            seconds=sign * duration.second,
        ),
    )


def parse_date_duration(date_str: str, sign: int) -> Duration:
    date_designators = OrderedDict(
        (("Y", "years"), ("M", "months"), ("D", "days"), ("W", "weeks"))
    )

    duration = DateDuration()
    tmp_value = ""

    for idx, ch in enumerate(date_str):
        if is_time(ch):
            if tmp_value != "" and tmp_value == date_str[:idx]:
                # PYYYY-MM-DDThh:mm:ss
                # PYYYYMMDDThhmmss
                return parse_datetime_duration(date_str, sign)

            time_idx = idx + 1
            time_str = date_str[time_idx:]

            if time_str == "":
                raise NoTime("Wanted time, no time provided")

            return Duration(duration, parse_time_duration(time_str, sign))

        if is_letter(ch):
            try:
                key = parse_designator(date_designators, ch)
                value = sign * Decimal(tmp_value)
            except OutOfDesignators as exc:
                raise IncorrectDesignator(
                    f"Wrong date designator, or designator in the wrong order: {ch}"
                ) from exc
            except InvalidOperation as exc:
                raise UnparseableValue(
                    f"Value could not be parsed as decimal: {tmp_value}"
                ) from exc

            if is_week(ch) and duration != DateDuration():
                raise IncorrectDesignator(
                    "Week is incompatible with any other date designator"
                )

            setattr(duration, key, value)
            tmp_value = ""

            continue

        if is_number(ch):
            if ch == ",":
                tmp_value += "."
            else:
                tmp_value += ch

            continue

        raise UnknownToken(f"Token not recognizable: {ch}")

    return Duration(duration, TimeDuration())


def parse_time_duration(time_str: str, sign: int) -> TimeDuration:
    time_designators = OrderedDict((("H", "hours"), ("M", "minutes"), ("S", "seconds")))

    duration = TimeDuration()
    tmp_value = ""

    for ch in time_str:
        if is_letter(ch):
            try:
                key = parse_designator(time_designators, ch)
                value = sign * Decimal(tmp_value)
            except OutOfDesignators as exc:
                raise IncorrectDesignator(
                    f"Wrong time designator, or designator in the wrong order: {ch}"
                ) from exc
            except InvalidOperation as exc:
                raise UnparseableValue(
                    f"Value could not be parsed as decimal: {tmp_value}"
                ) from exc

            setattr(duration, key, value)
            tmp_value = ""

            continue

        if is_number(ch):
            if ch == ",":
                tmp_value += "."
            else:
                tmp_value += ch

            continue

        raise UnknownToken(f"Token not recognizable: {ch}")

    return duration



from __future__ import annotations

from typing import TYPE_CHECKING

from isoduration.formatter.exceptions import DurationFormattingException

if TYPE_CHECKING:  # pragma: no cover
    from isoduration.types import DateDuration, Duration


def check_global_sign(duration: Duration) -> int:
    is_date_zero = (
        duration.date.years == 0
        and duration.date.months == 0
        and duration.date.days == 0
        and duration.date.weeks == 0
    )
    is_time_zero = (
        duration.time.hours == 0
        and duration.time.minutes == 0
        and duration.time.seconds == 0
    )

    is_date_negative = (
        duration.date.years <= 0
        and duration.date.months <= 0
        and duration.date.days <= 0
        and duration.date.weeks <= 0
    )
    is_time_negative = (
        duration.time.hours <= 0
        and duration.time.minutes <= 0
        and duration.time.seconds <= 0
    )

    if not is_date_zero and not is_time_zero:
        if is_date_negative and is_time_negative:
            return -1
    elif not is_date_zero:
        if is_date_negative:
            return -1
    elif not is_time_zero:
        if is_time_negative:
            return -1

    return +1


def validate_date_duration(date_duration: DateDuration) -> None:
    if date_duration.weeks:
        if date_duration.years or date_duration.months or date_duration.days:
            raise DurationFormattingException(
                "Weeks are incompatible with other date designators"
            )



from isoduration.formatter import format_duration
from isoduration.formatter.exceptions import DurationFormattingException
from isoduration.parser import parse_duration
from isoduration.parser.exceptions import DurationParsingException

__all__ = (
    "format_duration",
    "parse_duration",
    "DurationParsingException",
    "DurationFormattingException",
)



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal

from isoduration.formatter import format_duration
from isoduration.operations import add


@dataclass
class DateDuration:
    years: Decimal = Decimal(0)
    months: Decimal = Decimal(0)
    days: Decimal = Decimal(0)
    weeks: Decimal = Decimal(0)

    def __neg__(self) -> DateDuration:
        return DateDuration(
            years=-self.years,
            months=-self.months,
            days=-self.days,
            weeks=-self.weeks,
        )


@dataclass
class TimeDuration:
    hours: Decimal = Decimal(0)
    minutes: Decimal = Decimal(0)
    seconds: Decimal = Decimal(0)

    def __neg__(self) -> TimeDuration:
        return TimeDuration(
            hours=-self.hours,
            minutes=-self.minutes,
            seconds=-self.seconds,
        )


class Duration:
    def __init__(self, date_duration: DateDuration, time_duration: TimeDuration):
        self.date = date_duration
        self.time = time_duration

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.date}, {self.time})"

    def __str__(self) -> str:
        return format_duration(self)

    def __hash__(self) -> int:
        return hash(
            (
                self.date.years,
                self.date.months,
                self.date.days,
                self.date.weeks,
                self.time.hours,
                self.time.minutes,
                self.time.seconds,
            )
        )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Duration):
            return self.date == other.date and self.time == other.time

        raise NotImplementedError

    def __neg__(self) -> Duration:
        return Duration(-self.date, -self.time)

    def __add__(self, other: datetime) -> datetime:
        if isinstance(other, datetime):
            return add(other, self)

        raise NotImplementedError

    __radd__ = __add__

    def __sub__(self, other: object) -> None:
        raise NotImplementedError

    def __rsub__(self, other: datetime) -> datetime:
        if isinstance(other, datetime):
            return -self + other

        raise NotImplementedError



from decimal import ROUND_FLOOR, Decimal


def quot2(dividend: Decimal, divisor: Decimal) -> Decimal:
    return (dividend / divisor).to_integral_value(rounding=ROUND_FLOOR)


def mod2(dividend: Decimal, divisor: Decimal) -> Decimal:
    return dividend - quot2(dividend, divisor) * divisor


def quot3(value: Decimal, low: Decimal, high: Decimal) -> Decimal:
    dividend = value - low
    divisor = high - low
    return (dividend / divisor).to_integral_value(rounding=ROUND_FLOOR)


def mod3(value: Decimal, low: Decimal, high: Decimal) -> Decimal:
    dividend = value - low
    divisor = high - low
    return mod2(dividend, divisor) + low


def max_day_in_month(year: Decimal, month: Decimal) -> Decimal:
    norm_month = int(mod3(month, Decimal(1), Decimal(13)))
    norm_year = year + quot3(month, Decimal(1), Decimal(13))

    if norm_month in (1, 3, 5, 7, 8, 10, 12):
        return Decimal(31)
    if norm_month in (4, 6, 9, 11):
        return Decimal(30)

    is_leap_year = (
        mod2(norm_year, Decimal(400)) == 0
        or mod2(norm_year, Decimal(100)) != 0
        and mod2(norm_year, Decimal(4)) == 0
    )
    if norm_month == 2 and is_leap_year:
        return Decimal(29)

    return Decimal(28)



from isoduration.parser.exceptions import EmptyDuration
from isoduration.parser.parsing import parse_date_duration
from isoduration.parser.util import is_period
from isoduration.types import Duration


def parse_duration(duration_str: str) -> Duration:
    if len(duration_str) < 2:
        raise EmptyDuration("No duration information provided")

    beginning = 1
    first = duration_str[beginning - 1]

    sign = +1

    if first == "+":
        beginning += 1
    if first == "-":
        sign = -1
        beginning += 1

    prefix = duration_str[beginning - 1]
    duration = duration_str[beginning:]

    if not is_period(prefix):
        raise EmptyDuration("No prefix provided")

    return parse_date_duration(duration, sign)



from __future__ import annotations

from typing import TYPE_CHECKING

from isoduration.constants import PERIOD_PREFIX
from isoduration.formatter.checking import check_global_sign
from isoduration.formatter.formatting import format_date, format_time

if TYPE_CHECKING:  # pragma: no cover
    from isoduration.types import Duration


def format_duration(duration: Duration) -> str:
    global_sign = check_global_sign(duration)
    date_duration_str = format_date(duration.date, global_sign)
    time_duration_str = format_time(duration.time, global_sign)

    duration_str = f"{date_duration_str}{time_duration_str}"
    sign_str = "-" if global_sign < 0 else ""

    if duration_str == PERIOD_PREFIX:
        return f"{PERIOD_PREFIX}0D"

    return f"{sign_str}{duration_str}"
